/*
 * Copyright (c) 2001-2004 The Trustees of Indiana University.  
 *                         All rights reserved.
 * Copyright (c) 1998-2001 University of Notre Dame. 
 *                         All rights reserved.
 * Copyright (c) 1994-1998 The Ohio State University.  
 *                         All rights reserved.
 * 
 * This file is part of the LAM/MPI software package.  For license
 * information, see the LICENSE file in the top level directory of the
 * LAM/MPI source distribution.
 * 
 * $HEADER$
 *
 * $Id: ring.c,v 1.4 2002/10/09 20:55:30 brbarret Exp $
 *
 * Simple example of MPI showing the ring message pattern with a
 * special datatype (counts 0 for a block). Thanks to Marchand Aurelia
 * for reporting the bug with this special datatype
 */

#include <stdio.h>
#include "mpi.h"

#include "lamtest_error.h"

#define NUM_VALUES 16
/*
 * The general scheme of this program is to send a message around a
 * "ring" of N MPI processes.  Thank is, rank 0 sends to rank 1, rank
 * 1 sends to rank 2, ..., rank (N-1) sends back to rank 0. A special
 * indexed datatype with counts within a block as 0 has been created.
 */

/* Static functions */

static int check_results (int[], int[]);


int 
main(int argc, char **argv)
{
  int message;
  int rank, size, tag, to, from, i;
  int disps[3] = {0, 1, 4};
  int blocks[3] = {0, 2, 0};
  int values[NUM_VALUES], newvalues[NUM_VALUES];
  MPI_Datatype new_type;

  /* Start up MPI */

  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &size);
 
  /* Arbitrarily choose 201 to be our tag.  Calculate the */
  /* rank of the to process in the ring.  Use the modulus */
  /* operator so that the last process "wraps around" to rank */
  /* zero. */

  message = 0;
  tag = 201;
  to = (rank + 1) % size;
  from = (rank + size - 1) % size;

  /* Build the special datatype */

  MPI_Type_indexed(3, blocks, disps, MPI_INT, &new_type);
  MPI_Type_commit(&new_type);

  /* Fill in the array to be used for the special datatype */

  for (i = 0; i < NUM_VALUES; ++i) {
      values[i] = i;
      newvalues[i] = -1;
  }

  /* I am the root, I will seed the ring and stop the ring */

  if (rank == 0) {
    MPI_Send(values, 4, new_type, to, tag, MPI_COMM_WORLD); 
    MPI_Recv(newvalues, 4, new_type, from, tag, MPI_COMM_WORLD, 
	     MPI_STATUS_IGNORE);
    if (check_results(values, newvalues) != 0)
	lamtest_error(__FILE__, __LINE__, "Got incorrect receives on "
		      "rank %d\n", rank);
  }

  /* I am not the root, so I will just get the message from my
     previous neighbir and send it to the next neighbor  */

  else {
    MPI_Recv(newvalues, 4, new_type, from, tag, MPI_COMM_WORLD, 
	     MPI_STATUS_IGNORE);
   
    if (check_results(values, newvalues) != 0)
	lamtest_error(__FILE__, __LINE__, "Got incorrect receives on "
		      "rank %d", rank);
    MPI_Send(newvalues, 4, new_type, to, tag, MPI_COMM_WORLD);
  }
  
  /* Quit */

  MPI_Finalize();
  return 0;
}


int
check_results(int values[], int newvalues[]) 
{
  int i, mod_index;

  for (i = 0; i < NUM_VALUES; ++i) {

    mod_index = i % 4;

    switch (mod_index) {

    case 0:
    case 3:
	if (newvalues[i] != -1) {
	  return -1;
	}
      break;
      
    default:
	if (newvalues[i] != values[i]){
	  return -1;
	}
      break;
    }
  }
  return 0;
}
	
